# - This file is part of the RKWard project (https://rkward.kde.org).
# SPDX-FileCopyrightText: by Thomas Friedrichsmeier <thomas.friedrichsmeier@kdemail.net>
# SPDX-FileCopyrightText: by Meik Michalke <meik.michalke@hhu.de>
# SPDX-FileContributor: The RKWard Team <rkward@kde.org>
# SPDX-License-Identifier: GPL-2.0-or-later
#' RKWard Plugin Test Suite Framework.
#'
#' \tabular{ll}{
#' Package: \tab rkwardtests\cr
#' Type: \tab Package\cr
#' Version: \tab 0.5.7\cr
#' Date: \tab 2012-06-03\cr
#' Depends: \tab R (>= 2.9.0),methods\cr
#' Encoding: \tab UTF-8\cr
#' License: \tab GPL (>= 2)\cr
#' LazyLoad: \tab yes\cr
#' URL: \tab https://rkward.kde.org\cr
#' }
#'
#' @description
#' A set of functions, classes and methods to test plugins that were written for RKWard.
#'
#' @aliases rkwardtests-package rkwardtests
#' @name rkwardtests-package
#' @docType package
#' @title RKWard Plugin Test Suite Framework.
#' @author Thomas Friedrichsmeier, Meik Michalke
#' @keywords package
NULL
