/*- This file is part of the RKWard project (https://rkward.kde.org).
SPDX-FileCopyrightText: by Thomas Friedrichsmeier <thomas.friedrichsmeier@kdemail.net>
SPDX-FileContributor: The RKWard Team <rkward@kde.org>
SPDX-License-Identifier: GPL-2.0-or-later
*/
include("dist_common.js");

function getDistSpecifics() {
	return initDistSpecifics(i18n('Cauchy distribution'), 'cauchy', [ "location", "scale" ], [], continuous);
}
