<!DOCTYPE kpartgui>
<!--- This file is part of the RKWard project (https://rkward.kde.org).
SPDX-FileCopyrightText: by Thomas Friedrichsmeier <thomas.friedrichsmeier@kdemail.net>
SPDX-FileContributor: The RKWard Team <rkward@kde.org>
SPDX-License-Identifier: GPL-2.0-or-later
-->
<kpartgui name="rkward_katepluginintegration" version="710">
	<MenuBar>
		<Menu name="window"><text>&amp;Window</text>
			<Menu name="window_activate"><text>&amp;Activate</text></Menu>
		</Menu>
	</MenuBar>
</kpartgui>
