/*- This file is part of the RKWard project (https://rkward.kde.org).
SPDX-FileCopyrightText: by Thomas Friedrichsmeier <thomas.friedrichsmeier@kdemail.net>
SPDX-FileContributor: The RKWard Team <rkward@kde.org>
SPDX-License-Identifier: GPL-2.0-or-later
*/
function calculate() {
	echo('model = glm (' + getValue("model") + ', data=' + getValue("model.table") + ')\n');
	echo('labels = ' + getValue("model.labels") + ' \n');
	echo('result = anova (model)\n');
}

function printout() {
	echo('rk.header ("Simple Anova")\n');
	echo('rk.print (result)\n');
}
